-------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local addonName, L = ...;

--
-- Function to return the index for any undefined indexes
--
local function defaultValueIndex(table, index)
  return index;
end

-- Set the function for any undefined indexes
setmetatable(L, {__index = defaultValueIndex});

local NUM_INVITES_PER_PAGE = 10;

local daysInYear =
{
  {0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335},
  {0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334}
};

local pointMapping =
{
  ["BL"] = {"BOTTOMLEFT"},
  ["BC"] = {"CENTER", "BOTTOM"},
  ["BR"] = {"BOTTOMRIGHT"},
  ["CL"] = {"CENTER", "LEFT"},
  ["CC"] = {"CENTER"},
  ["CR"] = {"CENTER", "RIGHT"},
  ["TL"] = {"TOPLEFT"},
  ["TC"] = {"CENTER", "TOP"},
  ["TR"] = {"TOPRIGHT"},
};

-------------------------------------------------------------------------------
--                L  O  C  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to disable a Niggles character list frame and its descendants
--
local function characterListDisable(self)
  self:SetEnabled(false);
  return;
end

--
-- Function to enable a Niggles character list frame and its descendants
--
local function characterListEnable(self)
  self:SetEnabled(true);
  return;
end

--
-- Function to set the enabled state of a Niggles character list frame and
-- its descendants
--
local function characterListSetEnabled(self, enable)
  -- Local Variables
  local buttons;

  -- Save the enabled status
  self.isEnabled = enable;

  -- Enable/Disable the label
  self.label:SetFontObject(enable and self.label.normalFont or
    self.label.disableFont);

  -- Enable/Disable the scroll frame and its buttons
  self.scrollFrame.scrollBar:SetEnabled(enable);
  buttons = self.scrollFrame.buttons;
  if (buttons ~= nil) then
    for idx = 1, #buttons do
      buttons[idx]:SetEnabled(enable);
    end
  end
  self.scrollFrame.update();

  -- Enable/Disable the edit box and Add button
  self.edit:SetEnabled(enable);
  self.add:SetEnabled(enable);

  return;
end

-------------------------------------------------------------------------------
--               G  L  O  B  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--                A  D  D  O  N     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

--
-- NOTE: The ID for each continent is any (although typically the first) valid
--       map ID for that continent, since each continent can have several map
--       IDs. The ID should NEVER be changed once set.
--
L.continents =
{
  associative = {},
  indexed =
  {
    -- 1: ID
    -- 2: Name
    -- 3: Mask
    -- 4: Flight Spell ID
    {905, nil, 0x00000200,     -1}, -- Argus
    {619, nil, 0x00000100, 233368}, -- Broken Isles
    {572, nil, 0x00000080, 191645}, -- Draenor
    { 13, nil, 0x00000004,      0}, -- Eastern Kingdoms
    { 12, nil, 0x00000002,      0}, -- Kalimdor
    {876, nil, 0x00000400,     -1}, -- Kul Tiras
    {113, nil, 0x00000010,      0}, -- Northrend
    {101, nil, 0x00000008,      0}, -- Outland
    {424, nil, 0x00000040,      0}, -- Pandaria
    {948, nil, 0x00000020,      0}, -- The Maelstrom
    {875, nil, 0x00000800,     -1}, -- Zandalar
  },
  mapContinent = {},
};
setmetatable(L.continents,
{
  __index = function(table, key)
    return (key == "length" and L.continents.init() or nil);
  end
});

L.playerAuras =
{
  list   = {},
  update = true
};

-------------------------------------------------------------------------------
--                A  D  D  O  N     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Method to get info for a continent by its index
--
function L.continents.byIndex(idx)
  -- Initialise the continent info, if required
  if (next(L.continents.associative) == nil) then
    L.continents.init();
  end

  -- Return the info for the continent with the specified index
  return unpack(L.continents.indexed[idx] or {});
end

--
-- Method to get info for a continent from a map ID
--
function L.continents.byMapId(mapId)
  -- Local Variables
  local mapContinent = L.continents.mapContinent;

  -- Initialise the continent info, if required
  if (next(mapContinent) == nil) then
    L.continents.init();
  end

  -- Return the info for the continent with the specified index
  return unpack(mapContinent[mapId] or {});
end

--
-- Method to get info for the current continent the player is on
--
function L.continents.player()
  -- Get the map info for the continent of the player's current zone
  local mapId = C_Map.GetBestMapForUnit("player");

  -- Return the info for the continent
  return (mapId ~= nil and L.continents.byMapId(mapId) or nil);
end

--
-- Method to initialise the continent info
--
function L.continents.init()
  -- Local Variables
  local associative = L.continents.associative;
  local current;
  local currentIdx;
  local idx;
  local indexed = L.continents.indexed;
  local mapContinent = L.continents.mapContinent
  local mapId;
  local mapInfo;
  local maxMask = 0;
  local other;
  local parentInfo;

  -- Check if the continent info has already been initialised
  if (next(associative) ~= nil) then
    return #indexed;
  end

  -- Initialise an entry for maps with no continent
  other = {0, L["Other"], 0x00000001, -1};

  -- Get the names of the known continents
  for currentIdx, current in ipairs(indexed) do
    mapInfo                 = C_Map.GetMapInfo(current[1]);
    current[2]              = mapInfo.name;
    associative[current[2]] = current;
    maxMask                 = math.max(current[3], maxMask);
  end

  --- Initialise a list of each map's parent
  for idx = 1, 2000 do
    -- Get any map info for the ID
    mapInfo = C_Map.GetMapInfo(idx);
    if (mapInfo ~= nil) then
      -- Process the map's info
      if (mapInfo.mapType == Enum.UIMapType.Continent) then
        -- If the continent is unknown...
        if (associative[mapInfo.name] == nil) then
          -- ...add it to the object
          maxMask = bit.lshift(maxMask, 1);
          table.insert(indexed, {idx, mapInfo.name, maxMask});
          associative[mapInfo.name] = indexed[#indexed];
          mapContinent[idx]         = indexed[#indexed];
        end

        -- Map the ID to a conti
        mapContinent[idx] = associative[mapInfo.name];
      elseif (mapInfo.parentMapID ~= null) then
        -- Save the parent's map ID
        mapContinent[idx] = mapInfo.parentMapID;
      else
        -- Set the map's continent to 'Other'
        mapContinent[idx] = other;
      end
    end
  end

  -- Find the continent for each map
  for mapId, parentInfo in pairs(mapContinent) do
    while (type(parentInfo) == "number") do
      parentInfo = mapContinent[parentInfo];
    end
    mapContinent[mapId] =
      (parentInfo == nil and other or parentInfo);
  end

  -- Sort the indexed table by name
  table.sort(indexed,
    function(first, second) return (first[2] < second[2]); end);
  table.insert(indexed, other);

  return #indexed;
end

--
-- Function to (roughly) convert a date & time into a number of minutes
--
function L.dateToMinutes(date)
  -- Local Variables
  local _;

  -- If no date/time has been specified...
  if (date == nil) then
    -- ...use the current date/time
    date = C_Calendar.GetDate();
  end

  -- Work out the number of minutes
  return (((((date.year*365)+math.floor((date.year-1)/4)+
    daysInYear[((date.year-1)%4 == 0 and 2 or 1)][date.month]+
    date.monthDay)*24)+date.hour)*60)+date.minute;
end

--
-- Function to restore the position and size of a frame, if previously saved
-- using the specified name
--
function L.layoutRestore(name, frame, restoreSize)
  -- Local Variables
  local current;

  -- Check the specified arguments are valid and there is a layout for the name
  if ((type(name) == "string") and
      (name ~= "") and
      (frame ~= nil) and
      (NigglesDBC.layout[name] ~= nil)) then
    -- Assign info to more convenient variables
    current = NigglesDBC.layout[name];

    -- Set the frame's point(s)
    frame:ClearAllPoints();
    if (pointMapping[current.point] ~= nil) then
      for _, point in ipairs(pointMapping[current.point]) do
        frame:SetPoint(point, UIParent, (current.x or 0), (current.y or 0));
      end
    end

    -- Restore the frame's size, if required
    if (restoreSize) then
      frame:SetSize(math.max(current.width, 1), math.max(current.height, 1));
    end
  end

  return;
end

--
-- Function to save the position and size of a frame under the specified name
--
function L.layoutSave(name, frame, saveSize)
  -- Local Variables
  local botOffset;
  local centreH;
  local centreV;
  local leftOffset;
  local offset;
  local rightOffset;
  local topOffset;
  local uiCentreH;
  local uiCentreV;

  -- Check the specified arguments are valid
  if ((type(name) == "string") and (name ~= "") and (frame ~= nil)) then
    -- Get info for the specified frame
    leftOffset       = math.max(frame:GetLeft(), 0);
    rightOffset      = math.max(UIParent:GetWidth()-frame:GetRight(), 0);
    topOffset        = math.max(UIParent:GetHeight()-frame:GetTop(), 0);
    botOffset        = math.max(frame:GetBottom(), 0);
    centreH, centreV = frame:GetCenter();

    -- Get centre of the UIParent
    uiCentreH, uiCentreV = UIParent:GetCenter();

    -- Initialise/reset the layout info for the specified name
    if (type(NigglesDBC.layout) ~= "table") then
      NigglesDBC.layout = {};
    end
    NigglesDBC.layout[name] =
    {
      point  = "",
      x      = 0,
      y      = 0,
      width  = -1,
      height = -1,
    };
    current = NigglesDBC.layout[name];

    -- Save the vertical positioning
    offset = math.abs(uiCentreV-centreV);
    if ((offset < topOffset) and (offset < botOffset)) then
      current.point = "C";
      current.y      = centreV-uiCentreV;
    elseif (topOffset < botOffset) then
      current.point = "T";
      current.y      = -topOffset;
    else
      current.point = "B";
      current.y      = botOffset;
    end

    -- Save the horizontal positioning
    offset = math.abs(uiCentreH-centreH);
    if ((offset < leftOffset) and (offset < rightOffset)) then
      current.point = current.point.."C";
      current.x      = centreH-uiCentreH;
    elseif (leftOffset < rightOffset) then
      current.point = current.point.."L";
      current.x      = leftOffset;
    else
      current.point = current.point.."R";
      current.x      = -rightOffset;
    end

    -- Save the size, if required
    if (saveSize) then
      current.width, current.height = frame:GetSize();
    end
  end

  return;
end

--
-- Method to get the player's current list of auras.
--
-- NOTE: The returned list MUST be treated as read-only.
--
L.playerAuras.get = function()
  -- Local Variables
  local auraId;
  local auraIdx = 1;
  local auraName;
  local auras = L.playerAuras.list;

  -- Check if the list needs updating
  if (L.playerAuras.update) then
    -- Clear all existing auras
    wipe(auras);

    -- Add the current auras to the list
    auraName, auraId = UnitAura("player", auraIdx); 
    while (auraId ~= null) do
      auras[auraName] = true;
      auras[auraId  ] = true;
      auraIdx = auraIdx+1;
      auraName, auraId = UnitAura("player", auraIdx);
    end
  end

  return auras;
end

--
-- Method to check if the player currently has an aura
--
L.playerAuras.has = function(auraId)
  -- Local Variables
  local auras = L.playerAuras.list;

  -- Update the list of auras, if required
  if (L.playerAuras.update) then
    auras = L.playerAuras.get();
  end

  return (auras[auraId] == true);
end

--
-- Method called when the player's auras change. It simply sets a flag that
-- indicates the list of auras should be rebuilt.
--
L.playerAuras.onEvent = function()
  -- Set the flag that indicates the list should be updated
  L.playerAuras.update = true;

  return;
end

-------------------------------------------------------------------------------
--               G  L  O  B  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to initialise a Niggles Character List
--
function L.characterListInit(frame, label, buttonTemplate, updateFunc, addFunc)
  -- Save some of the arguments
  frame.buttonTemplate = buttonTemplate;
  frame.addFunc        = addFunc;

  -- Set the frame's label
  frame.label:SetText(label);
  frame.label.normalFont  = "GameFontNormal";
  frame.label.disableFont = "GameFontDisable";

  -- Initialise the scroll frame
  frame.scrollFrame.update = updateFunc;
  frame.scrollFrame.scrollBar.doNotHide = true;
  HybridScrollFrame_CreateButtons(frame.scrollFrame, buttonTemplate, 1, 0);

  -- Adjust some colours
  frame.scrollFrame.scrollBar.trackBG:SetVertexColor(0.2, 0.13, 0.08, 0.25);
  frame.border:SetBackdropBorderColor(0.75, 0.75, 0.75, 1.0);

  return;
end

--
-- Function to process the Enter key in an Character List edit box
--
function NigglesCharacterListEditOnEnterPressed(self)
  -- Local Variables
  local name;
  local frame = self:GetParent();

  -- Check the AutoComplete frame doesn't want the key press
  if (not AutoCompleteEditBox_OnEnterPressed(self)) then
    -- Get the name entered by the user
    name = strtrim(self:GetText());
    if (name ~= "") then
      -- Capitalize the name
      name = (name:sub(1,1):upper())..(name:sub(2):lower());

      -- Call the add function
      if (frame.addFunc ~= nil) then
        frame.addFunc(frame, name);
      end

      -- Clear the edit frame
      self:SetText("");
      AutoComplete_HideIfAttachedTo(self)
    else
      self:ClearFocus();
    end
  end

  return;
end

--
-- Function to tidy up when a Character List Edit frame loses focus
--
function NigglesCharacterListEditOnFocusLost(self)
  AutoCompleteEditBox_OnEditFocusLost(self);
  self:HighlightText(0, 0);
  local trimmedText = strtrim(self:GetText());
  if (trimmedText == "") then
    self:SetText("");
  end

  return;
end

--
-- Function called when a Niggles Character List scroll frame changes size
--
-- NOTE: The buttons are created when the frame changes size so the character
--       list doesn't have to be fixed
--
function NigglesCharacterListScrollFrameOnSize(self)
  -- Local Variables
  local parent = self:GetParent();
  local isEnabled;
  local buttons;

  -- Check a button template has been supplied
  if (parent.buttonTemplate ~= nil) then
    -- Create buttons, if required
    HybridScrollFrame_CreateButtons(self,
      self:GetParent().buttonTemplate, 1, 0);

    -- Enable/Disable the buttons
    for _, button in ipairs(self.buttons) do
      button:SetEnabled(parent.isEnabled);
    end

    -- Update the scroll frame, if possible
    if (self.update ~= nil) then
      self.update();
    end
  end

  return;
end

--
-- Function called when a Niggles Character List is created
--
function NigglesCharacterListOnLoad(self)
  -- Add some extra methods to the frame
  self.Disable    = characterListDisable;
  self.Enable     = characterListEnable;
  self.SetEnabled = characterListSetEnabled;

  return;
end

--
-- Function to enable/disable a frame and all its descendants. This function
-- is recursive.
--
function NigglesEnableFrame(frame, enable)
  -- Local Variables
  local childType;
  local frameType;

  -- Enable/disable the frame, if possible
  if ((frame.Enable) and (enable == true)) then
    frame:Enable();
  elseif ((frame.Disable) and (enable == false)) then
    frame:Disable();
  elseif (frame.GetObjectType) then
    frameType = frame:GetObjectType();
    if (frameType == "Frame") then
      -- Process each region
      for _, child in ipairs({frame:GetRegions()}) do
        -- Get the child's type
        childType = child:GetObjectType();

        -- Process the child, based on its type
        if (childType == "FontString") then
          if ((child.normalFont ~= nil) and (enable == true)) then
            child:SetFontObject(child.normalFont);
          elseif ((child.disableFont ~= nil) and (enable == false)) then
            child:SetFontObject(child.disableFont);
          end
        end
      end
    elseif (frameType == "EditBox") then
      frame:EnableJoystick(enable);
      frame:EnableMouse(enable);
      frame:EnableMouseWheel(enable);
      frame:EnableKeyboard(enable);
      frame:ClearFocus();
      if ((frame.normalFont ~= nil) and (enable == true)) then
        frame:SetFontObject(frame.normalFont);
      elseif ((frame.disableFont ~= nil) and (enable == false)) then
        frame:SetFontObject(frame.disableFont);
      end
    end
  end

  -- Process all children of the current frame
  for _, child in ipairs({frame:GetChildren()}) do
    NigglesEnableFrame(child, enable);
  end

  return;
end

--
-- Function to get the localized version of a string
--
function NigglesGetLocaleString(string)
  return L[string];
end

--
-- Function to get all the points set for a frame and store them in an array
--
function NigglesFramePointsGet(frame, array)
  -- Clear the array
  wipe(array);

  -- Get all the frame's points
  for pointIdx = 1, frame:GetNumPoints() do
    array[pointIdx] = {frame:GetPoint(pointIdx)};
  end

  return;
end

--
-- Function to set all the points specified in an array for a frame
--
function NigglesFramePointsSet(frame, array)
  -- Clear all the frame's points
  frame:ClearAllPoints();

  -- Set all the frame's points
  for pointIdx = 1, #array do
    frame:SetPoint(array[pointIdx][1], array[pointIdx][2], array[pointIdx][3],
      array[pointIdx][4], array[pointIdx][5]);
  end

  return;
end

--
-- Function to check if a value is in a table, ignoring case
--
function NigglesTableContains(table, value)
  -- Local Variables
  local idx = 1;

  -- Search the table for the specified value
  value = value:lower();
  while (table[idx]) do
    if (value == table[idx]:lower()) then
      return 1;
    end
    idx = idx+1;
  end

  return nil;
end

--
-- Function to set the contents of the Niggles tooltip. It clears any old lines
-- and then sets the new contents. It then checks if new lines have been added
-- to the tooltip. If so then the lines are added to the arrays used to scan
-- the tooltip's contents more efficiently.
--
-- NOTE: While the tooltip's normal 'Set...' methods could be hooked, there
--       are so many that it was deemed better to use this technique to keep
--       the arrays of text lines up to date.
--
function NigglesTooltipSetContents(self, funcName, ...)
  -- Local Variables
  local numLines;
  local numElements = #(self.textLeft);

  -- Clear any old content
  self:ClearLines();

  -- Set the tooltip's new contents
  (self[funcName])(self, ...);

  -- Check if the tooltip has new lines
  numLines = self:NumLines();
  if (numLines > numElements) then
    -- Add the new lines to the arrays used to scan the tooltip
    for lineIdx = numElements+1, numLines do
      self.textLeft[lineIdx]  = _G["NigglesTooltipTextLeft"..lineIdx];
      self.textRight[lineIdx] = _G["NigglesTooltipTextRight"..lineIdx];
    end
  end

  return numLines;
end

--
-- Function called when the Niggles tooltip is loaded. It sets the owner for
-- the tooltip, initialise the arrays used to scan the tooltip and adds a
-- method for setting the contents of the tooltip.
--
function NigglesTooltipOnLoad(self)
  -- Set the tooltip's owner
  self:SetOwner(WorldFrame, "ANCHOR_NONE");

  -- Initialise the arrays used to scan the tooltip's contents
  self.textLeft  = {_G["NigglesTooltipTextLeft1"]};
  self.textRight = {_G["NigglesTooltipTextRight1"]};

  -- Add a new method for setting the contents of the tooltip
  self.SetContents = NigglesTooltipSetContents;

  return;
end

--
-- Function to initialise the Niggles Warning popup
--
function NigglesWarningPopupInit(text, postscript, okLabel, cancelLabel, callback)
  -- Local Variables
  local popup;

  -- Create the Niggles warning popup, if required
  if (NigglesWarningPopup == nil) then
    CreateFrame("Frame", "NigglesWarningPopup", UIParent,
      "NigglesWarningPopupTemplate");
  end
  popup = NigglesWarningPopup;

  -- Set the warning popup's text
  popup.text:SetText(text);
  popup.postscript:SetShown((postscript ~= nil));
  if (postscript ~= nil) then
    popup.postscript:SetText(tostring(postscript));
  end

  -- Set the buttons up
  popup.buttons.ok:SetText(tostring(okLabel));
  popup.buttons.cancel:SetShown((cancelLabel ~= nil));
  if (cancelLabel ~= nil) then
    popup.buttons.cancel:SetText(tostring(cancelLabel));
    popup.buttons:SetWidth((popup.buttons.cancel:GetWidth()*2)+10);
  else
    popup.buttons:SetWidth(popup.buttons.ok:GetWidth());
  end
  if (type(callback) == "function") then
    popup.buttons.callback = callback;
  else
    popup.buttons.callback = StaticPopupSpecial_Hide;
  end

  -- Hide the optional parts of the popup
  popup.list:Hide();
  popup.item:Hide();

  return;
end

-------------------------------------------------------------------------------
--                 G  L  O  B  A  L     C  L  A  S  S  E  S
-------------------------------------------------------------------------------

--
-- Class to provide a history for values, such as achievements and recipes
--
NigglesHistoryClass =
{
  -- Class Variables
  maxElements = 64,

  -- Method to add an item to the history list
  Add = function(self, value)
    -- Check a value was specified
    if ((value ~= nil) and (value ~= "")) then
      -- Check if the list is empty
      if (self.base == 0) then
        self.base        = 1;
        self.current     = 1;
        self.top         = 1;
        self.elements[1] = value;
      elseif (self.elements[self.current] ~= value) then
        -- Update the current, top and base indices
        self.current = (self.current == self.maxElements and 1 or
          self.current+1);
        self.top = self.current;
        if (self.current == self.base) then
          self.base = (self.base == self.maxElements and 1 or self.base+1);
        end

        -- Add the value to the history
        self.elements[self.current] = value;
      end
    end

    return;
  end;

  --
  -- Method to clear the history
  --
  Clear = function(self)
    -- Set the base index to a special value to indicate the history is empty
    self.base = 0;
  end;

  --
  --
  -- Method to get the current value from the history list
  Get = function(self)
    return (self.base ~= 0 and self.elements[self.current] or nil);
  end;

  --
  -- Method to check if moving backward or forward is valid in the history
  --
  IsMoveValid = function(self, backward)
    -- Check if moving in the specified direction is valid
    if (backward) then
      return ((self.base ~= 0) and (self.current ~= self.base));
    else
      return ((self.base ~= 0) and (self.current ~= self.top));
    end
  end;

  --
  -- Method to move one element back or forward in the history list
  --
  Move = function(self, backward)
    -- Local Variables
    local value = nil;

    -- Check the history list isn't empty
    if (self.base ~= 0) then
      -- Move in the specified direction in the history list, if possible
      if ((backward) and (self.current ~= self.base)) then
        -- Decrement the current index
        self.current = (self.current == 1 and self.maxElements or
          self.current-1);
        value = self.elements[self.current];
      elseif ((not backward) and (self.current ~= self.top)) then
        -- Increment the current index
        self.current = (self.current == self.maxElements and
          1 or self.current+1);
        value = self.elements[self.current];
      end
    end

    -- Return the current recipe name
    return value;
  end;

  --
  -- Method to create a new class object
  --
  New = function(self, maxElements)
    -- Create a new table for the object
    local object =
    {
      base     = 0,
      top      = 0,
      current  = 0,
      elements = {},
    };

    -- Set the object's metatable to use the default class values
    setmetatable(object, self);
    self.__index = self;

    -- Save the maximum number of elements for the history, if specified
    if ((maxElements ~= nil) and
        (type(maxElements) == "number") and
        (maxElements > 1)) then
      object.maxElements = maxElements;
    end

    -- Return the new object
    return object;
  end;
};
