-------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local addonName, L = ...;

-- Shapeshift forms for classes that should ALWAYS have an active form
local classShapeshiftForm =
{
  ["DEATHKNIGHT"] = L["Presence"],
  ["MONK"]        = L["Stance"],
  ["PALADIN"]     = L["Seal"],
  ["WARRIOR"]     = L["Stance"],
};

local factionInfo =
{
  ["Alliance"] = 1,
  ["Horde"]    = 0,
};

local HELP_COLOR_CODE = "|cFF8080FF";

-- Valid commands for Niggles
local validCommands =
{
  {command = L["General"]:lower(),     category = "General"    },
  {command = L["Warnings"]:lower(),    category = "Warnings"   },
  {command = L["AutoLoot"]:lower(),    category = "AutoLoot"   },
  {command = L["Professions"]:lower(), category = "Professions"},
  {command = L["Raids"]:lower(),       category = "Raids"      },
  {command = L["Events"]:lower(),      category = "CalEvents"  },
  {command = L["Companions"]:lower(),  category = "Companions" },
  {command = "debug",                  func     = "dumpDebug"  },
};

-------------------------------------------------------------------------------
--                G  O  B  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

-- Labels for the key bindings
_G["BINDING_HEADER_NIGGLES"]                     = L["Niggles"];
_G["BINDING_NAME_CLICK NigglesMount:LeftButton"] = L["Mount"];
_G["BINDING_NAME_NIGGLES_TOGGLELOCATOR"]         = L["Toggle Locator"];

SLASH_NIGGLES1 = "/niggles";
SLASH_NIGGLES2 = "/ngl";

-------------------------------------------------------------------------------
--                L  O  C  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to process any post-combat callback functions
--
local function processPostCombat()
  -- Local Variables
  local postCombat = Niggles.postCombat;

  -- Call each callback function in the post-combat table
  for idx, callback in ipairs(postCombat) do
    -- Check the callback is a function
    if (type(callback) == "function") then
      -- Call the callback function
      callback();
    end

    -- Remove the callback from the post-combat table
    postCombat[idx] = nil;
  end

  return;
end

-------------------------------------------------------------------------------
--                A  D  D  O  N     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to dump debug info
--
function L.dumpDebug()
  -- Local Variables
  local info = ChatTypeInfo["SYSTEM"];

  -- Dump the debug info
  DEFAULT_CHAT_FRAME:AddMessage("Niggles Debug Info:", info.r, info.g, info.b);
  L.equipmentDumpDebug();

  return;
end

-------------------------------------------------------------------------------
--               G  L  O  B  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to process the slash command
--
function SlashCmdList.NIGGLES(message, editbox)
  -- Local Variables
  local command, arguments = message:match("^(%S*)%s*(.-)$");
  local category = nil
  local validCmd;

  -- Check a command was specified
  if ((command ~= nil) and (command ~= "")) then
    -- Search the valid commands for one that matches the specified command
    for _, cmdInfo in ipairs(validCommands) do
      if (cmdInfo.command:find(command, 1, true) == 1) then
        validCmd = cmdInfo;
        break;
      end
    end
  end

  -- Check if the command was found
  if ((validCmd ~= nil) and (validCmd.category ~= nil)) then
    -- TODO: Check if the Blizzard bug is fixed that means the first time the
    --       Interface panel doesn't switch to the specified category
    InterfaceOptionsFrame_OpenToCategory(
      _G["NigglesOptions"..validCmd.category]);
    InterfaceOptionsFrame_OpenToCategory(
      _G["NigglesOptions"..validCmd.category]);
  elseif ((validCmd ~= nil) and 
          (validCmd.func ~= nil) and 
          (type(L[validCmd.func]) == "function")) then
    (L[validCmd.func])();
  else
    -- Display the help
    DEFAULT_CHAT_FRAME:AddMessage(HELP_COLOR_CODE.."/niggles||/ngl");
    for _, cmdInfo in ipairs(validCommands) do
      DEFAULT_CHAT_FRAME:AddMessage(HELP_COLOR_CODE..
        string.gsub(cmdInfo.command, "(.)(.*)", "  %1[%2]"));
    end
  end

  return;
end

--
-- Function to process events
--
function NigglesOnEvent(self, event, ...)
  -- Process the event that occurred
  if (event == "UNIT_SPELLCAST_SUCCEEDED") then
    -- Get the unit and spell IDs
    local unitId, _, spellId = ...;

    -- Check the spell is 'Growl' from a friendly pet in a multi-player party
    if ((spellId == 2649) and
        (UnitIsFriend("player", unitId)) and
        (GetNumGroupMembers() > 1)) then
      -- Check the player is in a PvE instance
      local _, instanceType = IsInInstance();
      if ((instanceType == "party") or
          (instanceType == "raid")) then
        Niggles:AddWarning(L["Pet Growled"], UnitName(unitId));
        PlaySound(SOUNDKIT.RAID_WARNING);
      end
    end
  elseif (event == "UNIT_AURA") then
    L.playerAuras.onEvent();
  elseif (event == "PLAYER_REGEN_DISABLED") then
    -- Check if the player is in a vehicle
    if (not UnitInVehicle("player")) then
      -- Local Variables
      local playSound = false;
      local numWarnings = 0;

      -- Get the ID of the weapon in the player's main hand
      local itemId = GetInventoryItemID("player", self.mainHandSlotId);
      if (itemId) then
        -- Display a warning if a lance or fishing rod is equipped
        if ((NigglesDBC.settings.WarningLance) and
            (L.equipmentIsLance(itemId))) then
          Niggles:AddWarning(L["Lance Equipped"]);
          numWarnings = numWarnings+1;
        elseif ((NigglesDBC.settings.WarningRod) and
                (IsEquippedItemType(L["Fishing Pole"]))) then
          Niggles:AddWarning(L["Fishing Pole Equipped"]);
          numWarnings = numWarnings+1;
        end
      end

      -- Display a warning if there are low level items equipped or empty slots
      if (NigglesDBC.settings.WarningLowLevelItems) then
        if ((L.equipmentGetLowIlvl() > 0) and
            (numWarnings == 0)) then
          Niggles:AddWarning(L["Low Level Items Equipped"]);
          numWarnings = numWarnings+1;
        elseif ((L.equipmentGetEmptySlots() > 0) and
                (UnitLevel("player") >= 60)) then
          Niggles:AddWarning(L["Empty Equipment Slots"]);
          numWarnings = numWarnings+1;
        end
      end

      -- Play a sound to alert the user of any messages
      if (numWarnings > 0) then
        PlaySound(SOUNDKIT.RAID_WARNING);
      end
    end
  elseif (event == "PLAYER_REGEN_ENABLED") then
    processPostCombat();
  elseif (event == "GROUP_ROSTER_UPDATE") then
    -- Check the player isn't in a Raid Finder raid
    local _, _, difficulty = GetInstanceInfo();
    if ((difficulty == nil) or
        ((difficulty ~= 7) and (difficulty ~= 17))) then
      if ((NigglesDBC.settings.RaidsAutoPromote == true) and
          (UnitIsGroupLeader("PLAYER"))) then
        L.raidRosterPromoteAssists();
      end
      if (NigglesDBC.settings.CalEventsOnline) then
        NigglesCalEventOnUpdate();
      end
    end
    Niggles.inGroup = IsInGroup();
    Niggles.inRaid  = IsInRaid();
  elseif (event == "LOOT_READY") then
    local autoLoot = ...;
    if (autoLoot == false) then
      L.autoLoot();
    end
  elseif (event == "UNIT_CONNECTION") then
    NigglesCalEventOnUpdate();
  elseif(event == "PLAYER_EQUIPMENT_CHANGED") then
    L.equipmentScan("Character");
  elseif (event == "PET_JOURNAL_LIST_UPDATE") then
    C_Timer.After(1.0, NigglesPetsGetOwned);
  elseif (event == "PET_BATTLE_LEVEL_CHANGED") then
    local activePlayer, petSlot, petLevel = ...;
    if (activePlayer == LE_BATTLE_PET_ALLY) then
      NigglesPetUpdateLevel(petSlot, petLevel);
    end
  elseif (event == "UPDATE_SUMMONPETS_ACTION") then
    -- NOTE: This event appears to be triggered only when the player enters
    --       the world or when a pet is upgraded using a battle stone. If
    --       that ever changes, another method of detecting pets being
    --       upgraded will have to be used.
    NigglesPetRefreshOwned();
  elseif (event == "BAG_UPDATE") then
    local bagId = select(1, ...);
    if ((bagId >=0) and (bagId <= NUM_BAG_SLOTS)) then
      NigglesPetsBagUpdate();
    end
  elseif ((event == "QUEST_DETAIL") or
          (event == "QUEST_PROGRESS") or
          (event == "QUEST_COMPLETE")) then
    L.questsItemInfoUpdate();
  elseif (event == "LEARNED_SPELL_IN_TAB") then
    NigglesMountsSpellsUpdate();
  elseif ((event == "CALENDAR_UPDATE_EVENT") or
          (event == "CALENDAR_UPDATE_INVITE_LIST")) then
    NigglesCalEventOnUpdate();
  elseif (event == "CURRENCY_DISPLAY_UPDATE") then
    NigglesCurrencyCheck();
  elseif (event == "ITEM_LOCKED") then
    local containerId, slotId = select(1, ...);
    L.equipmentSaveItemInfo(containerId, slotId);
  elseif (event == "ITEM_UNLOCKED") then
    L.equipmentSaveItemInfo(nil, nil);
  elseif (event == "PET_BATTLE_OPENING_START") then
    L.equipmentCheckPetBattleXPBuffs();
  elseif (event == "ADDON_LOADED") then
    local addonName = select(1, ...);
    if (addonName == "Niggles") then
      NigglesOnLoaded();
    elseif (addonName == "Blizzard_AchievementUI") then
      NigglesAchievementsOnLoad();
    elseif (addonName == "Blizzard_ArchaeologyUI") then
      L.archaeologyOnLoad();
    elseif (addonName == "Blizzard_Calendar") then
      NigglesCalendarOnLoad();
    elseif (addonName == "Blizzard_Collections") then
      NigglesCompanionsOnLoad();
    elseif (addonName == "Blizzard_FlightMap") then
      L.flightListOnLoad();
    elseif (addonName == "Blizzard_InspectUI") then
      L.equipmentInspectUIOnLoad();
    elseif (addonName == "Blizzard_TradeSkillUI") then
      L.tradeSkillOnLoad();
    end
  elseif (event == "PLAYER_LOGIN") then
    -- Initialise features
    NigglesAchievementsInitialise();
    L.archaeologyInitialise();
    NigglesCompanionsInit();
    L.equipmentInit();
    NigglesMerchantInitialise();
    L.flightListInit();

    -- Show/Hide the Locator frame
    NigglesLocatorShow(NigglesDBC.settings.LocatorShown);

    -- Register for events
    NigglesRegisterForEvents();

    -- Check if the player is in a group and/or raid
    Niggles.inGroup = IsInGroup();
    Niggles.inRaid  = IsInRaid();
  end

  return;
end

--
-- Function to initialise the addon
--
function NigglesOnLoad(self)
  -- Local Variables
  local _;

  -- Initialise some variables
  self.updatedOptions  = nil;
  self.petMask         = 0;
  self.mainHandSlotId  = GetInventorySlotInfo("MainHandSlot");
  self.enabledLogging  = {false, false};
  self.enteredWorld    = false;
  self.postCombat      = {};

  -- Get the player's class, race and form
  _, self.playerClass = UnitClass("player");
  _, self.playerRace  = UnitRace("player");
  self.playerForm     = classShapeshiftForm[self.playerClass];
  self.factionId      = factionInfo[UnitFactionGroup("player")];

  -- Register the event triggered by the addon being loaded
  self:RegisterEvent("ADDON_LOADED");

  -- Add a method to display warnings
  function self:AddWarning(formatStr, ...)
    self:AddMessage(format(formatStr, ...), 1.0, 1.0, 0.0, 1.0);
    return;
  end

  return;
end

--
-- Function called once the addon has been loaded
--
function NigglesOnLoaded()
  -- Initialise the settings
  NigglesSettingsInit();

  -- Hook the standard UI
  NigglesHookUi();

  -- Register for some events
  Niggles:RegisterEvent("PLAYER_LOGIN");
  Niggles:RegisterEvent("PLAYER_LOGOUT");
  Niggles:RegisterEvent("LEARNED_SPELL_IN_TAB");
  Niggles:RegisterEvent("PLAYER_REGEN_ENABLED");
  Niggles:RegisterUnitEvent("UNIT_AURA", "PLAYER");

  -- Create the Niggles menu
  Niggles.menu = NigglesMenuClass:New();

  return;
end

--
-- Function to register/un-register for events, based on the options
--
function NigglesRegisterForEvents()
  -- Local Variables
  local settings = NigglesDBC.settings;

  -- Enable/Disable the Archaeology features
  L.archaeologyEnable();

  -- Enable/Disable the Achievement features
  NigglesAchievementsEnable();

  -- Check if changes in the group roster should be monitored
  if ((settings.RaidsAutoPromote) or
      (settings.CalEventsOnline)) then
    Niggles:RegisterEvent("GROUP_ROSTER_UPDATE");
  else
    Niggles:UnregisterEvent("GROUP_ROSTER_UPDATE");
  end
  if (settings.CalEventsOnline) then
    Niggles:RegisterEvent("UNIT_CONNECTION");
  else
    Niggles:UnregisterEvent("UNIT_CONNECTION");
  end

  -- Check if warnings should be generated for pets growling
  if (settings.WarningPetGrowl) then
    Niggles:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED");
  else
    Niggles:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED");
  end

  -- Check if warnings should be generated when entering combat
  if ((settings.WarningRod) or
      (settings.WarningLance) or
      (settings.WarningLowLevelItems) or
      (settings.WarningNoForm)) then
    Niggles:RegisterEvent("PLAYER_REGEN_DISABLED");
  else
    Niggles:UnregisterEvent("PLAYER_REGEN_DISABLED");
  end

  -- Check if warnings should be generated for unequipped pet battle items
  if (settings.WarningPetBattle) then
    Niggles:RegisterEvent("PET_BATTLE_OPENING_START");
  else
    Niggles:UnregisterEvent("PET_BATTLE_OPENING_START");
  end

  -- Check if warnings should be generated for currencies
  if (settings.WarningCurrency) then
    Niggles:RegisterEvent("CURRENCY_DISPLAY_UPDATE");
    NigglesCurrencyCheck();
  else
    Niggles:UnregisterEvent("CURRENCY_DISPLAY_UPDATE");
  end

  -- Check if warnings should be generated for items being sold
  if ((NigglesDBC.settings.WarningEquipSetItem) or
      (NigglesDBC.settings.WarningUncollectedSold)) then
    Niggles:RegisterEvent("ITEM_LOCKED");
    Niggles:RegisterEvent("ITEM_UNLOCKED");
  else
    Niggles:UnregisterEvent("ITEM_LOCKED");
    Niggles:UnregisterEvent("ITEM_UNLOCKED");
  end

  -- Check if updates to calendar events should be processed
  if ((settings.CalEventsOnline) or
      (settings.CalEventsRoles)) then
    Niggles:RegisterEvent("CALENDAR_UPDATE_EVENT");
    Niggles:RegisterEvent("CALENDAR_UPDATE_INVITE_LIST");
  else
    Niggles:UnregisterEvent("CALENDAR_UPDATE_EVENT");
    Niggles:UnregisterEvent("CALENDAR_UPDATE_INVITE_LIST");
  end

  -- Check if looting should be monitored
  if ((settings.AutoLootArch    ) or
      (settings.AutoLootCurrency) or
      (settings.AutoLootFish    ) or
      (settings.AutoLootMoney   )) then
    Niggles:RegisterEvent("LOOT_READY");
  else
    Niggles:UnregisterEvent("LOOT_READY");
  end

  -- Enable/Disable the Calendar features
  NigglesCalendarEnable();

  -- Enable logging if required or disable if previously enabled
  if ((settings.LogChat) or
      (Niggles.enabledLogging[1] == true)) then
    Niggles.enabledLogging[1] = settings.LogChat;
    LoggingChat(settings.LogChat);
  end
  if ((settings.LogCombat) or
      (Niggles.enabledLogging[2] == true)) then
    Niggles.enabledLogging[2] = settings.LogCombat;
    LoggingCombat(settings.LogCombat);
  end

  -- Set the font for the mail system
  if (settings.UiMailFont == true) then
    OpenMailBodyText:SetFontObject("QuestFont");
    SendMailBodyEditBox:SetFontObject("QuestFont");
  else
    OpenMailBodyText:SetFontObject("MailTextFontNormal");
    SendMailBodyEditBox:SetFontObject("MailTextFontNormal");
  end

  -- Enable/Disable features
  L.archaeologyEnable();
  L.questsEnable();
  L.equipmentEnable();
  L.mapLocationEnable();
  L.flightListEnable();
  NigglesCurrencyEnable();
  NigglesMerchantEnable();
  NigglesHookUiEnable();
  NigglesCompanionsEnable();
  L.tradeSkillEnable();

  return;
end
